#include <float.h>
#include <math.h>
#include "mex.h"

#define IA   16807
#define IM   2147483647
#define AM   (1.0 / IM)
#define IQ   127773
#define IR   2836
#define NTAB 32
#define NDIV (1 + (IM - 1) / NTAB)

#define NEIGHBOR_LEFT  0
#define NEIGHBOR_RIGHT 1

double ran1(long *idum, long *iy, long *iv)
{
    int j;
    long k;

    if ((*idum <= 0) || !*iy)
    {
        *idum = (-*idum < 1) ? 1 : -*idum;
        for (j = NTAB + 7; j >= 0; j--)
        {
            k = *idum / IQ;
            *idum = IA * (*idum - k * IQ) - IR * k;
            if (*idum < 0)
                *idum += IM;
            if (j < NTAB)
                iv[j] = *idum;
        }
        *iy = iv[0];
    }
    k = *idum / IQ;
    *idum = IA * (*idum - k * IQ) - IR * k;
    if (*idum < 0)
        *idum += IM;
    j = *iy / NDIV;
    *iy = iv[j];
    iv[j] = *idum;
    return AM * *iy;
}

double gasdev(long *idum, long *iy, long *iv, int *iset, double *gset)
{
    double fac, rsq, v1, v2;

    if (*idum < 0)
        *iset = 0;
    if (*iset == 0)
    {
        do
        {
            v1 = 2.0 * ran1(idum, iy, iv) - 1.0;
            v2 = 2.0 * ran1(idum, iy, iv) - 1.0;
            rsq = v1 * v1 + v2 * v2;
        }
        while ((rsq >= 1.0) || (rsq == 0.0));
        fac = sqrt(-2.0 * log(rsq) / rsq);
        *gset = v1 * fac;
        *iset = 1;
        return v2 * fac;
    }
    else
    {
        *iset = 0;
        return *gset;
    }
}

int get_neighbor(int neighbor, int x, int n)
{
    if (neighbor == NEIGHBOR_LEFT)
        return (x == 0) ? (n - 1) : (x - 1);
    return (x == n - 1) ? 0 : (x + 1);
}

void init_strategies(int *strategies, int *sum_strategies, double *initial_strategies, int n)
{
    int x;

    *sum_strategies = 0;
    for (x = 0; x < n; x++)
    {
        *(strategies + x) = (int)*(initial_strategies + x);
        *sum_strategies += *(strategies + x);
    }
}

void init_profits_per_strategy(double *profits_per_strategy, int *n_firms_per_strategy, int n_strategies)
{
    int i;

    for (i = 0; i < n_strategies; i++)
    {
        *(profits_per_strategy + i) = 0;
        *(n_firms_per_strategy + i) = 0;
    }
}

double calc_market_share(int *strategies, int x, int n, double price_min, double price_step_size)
{
    double price, price_neighbor_left, price_neighbor_right;
    int x_neighbor_left, x_neighbor_right;

    price = price_min + *(strategies + x) * price_step_size;
    x_neighbor_left = get_neighbor(NEIGHBOR_LEFT, x, n);
    price_neighbor_left = price_min + *(strategies + x_neighbor_left) * price_step_size;
    x_neighbor_right = get_neighbor(NEIGHBOR_RIGHT, x, n);
    price_neighbor_right = price_min + *(strategies + x_neighbor_right) * price_step_size;
    return (price_neighbor_left - price + 1) / 2 + (price_neighbor_right - price + 1) / 2;
}

void init_market_shares(double *market_shares, int *strategies, int n, double price_min, double price_step_size)
{
    int x;

    for (x = 0; x < n; x++)
        *(market_shares + x) = calc_market_share(strategies, x, n, price_min, price_step_size);
}

void calc_profits(double *profits, int *strategies, double *market_shares, int n, double price_min, double price_step_size)
{
    double price;
    int x;

    for (x = 0; x < n; x++)
    {
        price = price_min + *(strategies + x) * price_step_size;
        *(profits + x) = price * *(market_shares + x);
    }
}

void update_market_shares(double *market_shares, int *strategies, int x, int n, double price_min, double price_step_size)
{
    int x_neighbor;

    *(market_shares + x) = calc_market_share(strategies, x, n, price_min, price_step_size);
    x_neighbor = get_neighbor(NEIGHBOR_LEFT, x, n);
    *(market_shares + x_neighbor) = calc_market_share(strategies, x_neighbor, n, price_min, price_step_size);
    x_neighbor = get_neighbor(NEIGHBOR_RIGHT, x, n);
    *(market_shares + x_neighbor) = calc_market_share(strategies, x_neighbor, n, price_min, price_step_size);
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    double *initial_strategies, *market_shares, *mean_strategy, *profits, *profits_per_strategy;
    double experimentation_prob, gset, mean_profit_new_strategy, mean_profit_strategy, noise_level, price_min, price_step_size;
    int *n_firms_per_strategy, *strategies;
    int i, iset, j, learning_neighborhood, n, new_strategy, n_rounds, n_strategies, strategy, strategy_neighbor, sum_strategies, x, x_neighbor;
    long *iv;
    long idum, iy, random_seed;
    mwSize dims[2];

    initial_strategies = mxGetPr(prhs[0]);
    n = (int)(mxGetM(prhs[0]) * mxGetN(prhs[0]));
    price_min = mxGetScalar(prhs[1]);
    price_step_size = mxGetScalar(prhs[2]);
    n_strategies = (int)mxGetScalar(prhs[3]);
    learning_neighborhood = (int)mxGetScalar(prhs[4]);
    noise_level = mxGetScalar(prhs[5]);
    experimentation_prob = mxGetScalar(prhs[6]);
    n_rounds = (int)mxGetScalar(prhs[7]);
    random_seed = (long)mxGetScalar(prhs[8]);

    dims[0] = n;
    dims[1] = 1;
    plhs[0] = mxCreateNumericArray(2, dims, mxINT32_CLASS, mxREAL);
    strategies = (int*)mxGetPr(plhs[0]);
    plhs[1] = mxCreateDoubleMatrix(n_rounds, 1, mxREAL);
    mean_strategy = mxGetPr(plhs[1]);

    idum = -labs(random_seed);
    iy = 0;
    iv = mxMalloc(NTAB * sizeof(*iv));
    iset = 0;

    init_strategies(strategies, &sum_strategies, initial_strategies, n);
    market_shares = mxMalloc(n * sizeof(*market_shares));
    init_market_shares(market_shares, strategies, n, price_min, price_step_size);
    profits = mxMalloc(n * sizeof(*profits));
    profits_per_strategy = mxMalloc(n_strategies * sizeof(*profits_per_strategy));
    n_firms_per_strategy = mxMalloc(n_strategies * sizeof(*n_firms_per_strategy));

    for (i = 0; i < n_rounds; i++)
    {
        calc_profits(profits, strategies, market_shares, n, price_min, price_step_size);

        init_profits_per_strategy(profits_per_strategy, n_firms_per_strategy, n_strategies);
        x = (int)(n * ran1(&idum, &iy, iv)) % n;
        strategy = *(strategies + x);
        *(profits_per_strategy + strategy) += *(profits + x);
        (*(n_firms_per_strategy + strategy))++;
        x_neighbor = x;
        for (j = 0; j < learning_neighborhood; j++)
        {
            x_neighbor = get_neighbor(NEIGHBOR_LEFT, x_neighbor, n);
            strategy_neighbor = *(strategies + x_neighbor);
            *(profits_per_strategy + strategy_neighbor) += *(profits + x_neighbor) + noise_level * gasdev(&idum, &iy, iv, &iset, &gset);
            (*(n_firms_per_strategy + strategy_neighbor))++;
        }
        x_neighbor = x;
        for (j = 0; j < learning_neighborhood; j++)
        {
            x_neighbor = get_neighbor(NEIGHBOR_RIGHT, x_neighbor, n);
            strategy_neighbor = *(strategies + x_neighbor);
            *(profits_per_strategy + strategy_neighbor) += *(profits + x_neighbor) + noise_level * gasdev(&idum, &iy, iv, &iset, &gset);
            (*(n_firms_per_strategy + strategy_neighbor))++;
        }
        mean_profit_new_strategy = -DBL_MAX;
        for (j = 0; j < n_strategies; j++)
        {
            mean_profit_strategy = (*(n_firms_per_strategy + j) > 0) ? (*(profits_per_strategy + j) / *(n_firms_per_strategy + j) + 1e-10 * ran1(&idum, &iy, iv)) : -DBL_MAX;
            if (mean_profit_strategy > mean_profit_new_strategy)
            {
                new_strategy = j;
                mean_profit_new_strategy = mean_profit_strategy;
            }
        }
        if (new_strategy != strategy)
        {
            *(strategies + x) = new_strategy;
            sum_strategies += new_strategy - strategy;
            update_market_shares(market_shares, strategies, x, n, price_min, price_step_size);
        }

        for (x = 0; x < n; x++)
            if (ran1(&idum, &iy, iv) < experimentation_prob)
                if (ran1(&idum, &iy, iv) < 0.5)
                {
                    if (*(strategies + x) > 0)
                    {
                        (*(strategies + x))--;
                        sum_strategies--;
                        update_market_shares(market_shares, strategies, x, n, price_min, price_step_size);
                    }
                }
                else if (*(strategies + x) < n_strategies - 1)
                {
                    (*(strategies + x))++;
                    sum_strategies++;
                    update_market_shares(market_shares, strategies, x, n, price_min, price_step_size);
                }

        *(mean_strategy + i) = (double)sum_strategies / n;
    }

    mxFree(iv);
    mxFree(market_shares);
    mxFree(profits);
    mxFree(profits_per_strategy);
    mxFree(n_firms_per_strategy);
}
