#include <float.h>
#include <math.h>
#include "mex.h"

#define IA   16807
#define IM   2147483647
#define AM   (1.0 / IM)
#define IQ   127773
#define IR   2836
#define NTAB 32
#define NDIV (1 + (IM - 1) / NTAB)

#define MODEL_VARIANT_A 0
#define MODEL_VARIANT_B 1

#define SMALL_LEARNING_NEIGHBORHOOD 0
#define LARGE_LEARNING_NEIGHBORHOOD 1

#define N_NEIGHBORS         8
#define NEIGHBOR_NORTH      0
#define NEIGHBOR_NORTH_EAST 1
#define NEIGHBOR_EAST       2
#define NEIGHBOR_SOUTH_EAST 3
#define NEIGHBOR_SOUTH      4
#define NEIGHBOR_SOUTH_WEST 5
#define NEIGHBOR_WEST       6
#define NEIGHBOR_NORTH_WEST 7

double ran1(long *idum, long *iy, long *iv)
{
    int j;
    long k;

    if ((*idum <= 0) || !*iy)
    {
        *idum = (-*idum < 1) ? 1 : -*idum;
        for (j = NTAB + 7; j >= 0; j--)
        {
            k = *idum / IQ;
            *idum = IA * (*idum - k * IQ) - IR * k;
            if (*idum < 0)
                *idum += IM;
            if (j < NTAB)
                iv[j] = *idum;
        }
        *iy = iv[0];
    }
    k = *idum / IQ;
    *idum = IA * (*idum - k * IQ) - IR * k;
    if (*idum < 0)
        *idum += IM;
    j = *iy / NDIV;
    *iy = iv[j];
    iv[j] = *idum;
    return AM * *iy;
}

double gasdev(long *idum, long *iy, long *iv, int *iset, double *gset)
{
    double fac, rsq, v1, v2;

    if (*idum < 0)
        *iset = 0;
    if (*iset == 0)
    {
        do
        {
            v1 = 2.0 * ran1(idum, iy, iv) - 1.0;
            v2 = 2.0 * ran1(idum, iy, iv) - 1.0;
            rsq = v1 * v1 + v2 * v2;
        }
        while ((rsq >= 1.0) || (rsq == 0.0));
        fac = sqrt(-2.0 * log(rsq) / rsq);
        *gset = v1 * fac;
        *iset = 1;
        return v2 * fac;
    }
    else
    {
        *iset = 0;
        return *gset;
    }
}

int get_x_neighbor(int neighbor, int x, int m)
{
    if ((neighbor == NEIGHBOR_SOUTH_WEST) || (neighbor == NEIGHBOR_WEST) || (neighbor == NEIGHBOR_NORTH_WEST))
        return (x == 0) ? (m - 1) : (x - 1);
    if ((neighbor == NEIGHBOR_NORTH_EAST) || (neighbor == NEIGHBOR_EAST) || (neighbor == NEIGHBOR_SOUTH_EAST))
        return (x == m - 1) ? 0 : (x + 1);
    return x;
}

int get_y_neighbor(int neighbor, int y, int n)
{
    if ((neighbor == NEIGHBOR_NORTH) || (neighbor == NEIGHBOR_NORTH_EAST) || (neighbor == NEIGHBOR_NORTH_WEST))
        return (y == 0) ? (n - 1) : (y - 1);
    if ((neighbor == NEIGHBOR_SOUTH_EAST) || (neighbor == NEIGHBOR_SOUTH) || (neighbor == NEIGHBOR_SOUTH_WEST))
        return (y == n - 1) ? 0 : (y + 1);
    return y;
}

void init_strategies(int *strategies, int *sum_strategies, double *initial_strategies, int m, int n)
{
    int x, y;

    *sum_strategies = 0;
    for (x = 0; x < m; x++)
        for (y = 0; y < n; y++)
        {
            *(strategies + x * n + y) = (int)*(initial_strategies + x * n + y);
            *sum_strategies += *(strategies + x * n + y);
        }
}

void init_profits_per_strategy(double *profits_per_strategy, int *n_firms_per_strategy, int n_strategies)
{
    int i;

    for (i = 0; i < n_strategies; i++)
    {
        *(profits_per_strategy + i) = 0;
        *(n_firms_per_strategy + i) = 0;
    }
}

double calc_market_share_variant_A(int *strategies, int x, int y, int m, int n, double price_min, double price_step_size)
{
    double market_share, price, price_neighbor;
    int neighbor, x_neighbor, y_neighbor;

    price = price_min + *(strategies + x * n + y) * price_step_size;
    market_share = 0;
    for (neighbor = NEIGHBOR_NORTH; neighbor <= NEIGHBOR_WEST; neighbor += 2)
    {
        x_neighbor = get_x_neighbor(neighbor, x, m);
        y_neighbor = get_y_neighbor(neighbor, y, n);
        price_neighbor = price_min + *(strategies + x_neighbor * n + y_neighbor) * price_step_size;
        market_share += (price_neighbor - price + 1) / 2;
    }
    return market_share;
}

double calc_market_share_variant_B(int *strategies, int x, int y, int m, int n, double price_min, double price_step_size)
{
    double indifference_distances[N_NEIGHBORS];
    double a, market_share, price, price_neighbor;
    int neighbor1, neighbor2, neighbor3, x_neighbor, y_neighbor;

    price = price_min + *(strategies + x * n + y) * price_step_size;
    for (neighbor1 = NEIGHBOR_NORTH; neighbor1 <= NEIGHBOR_WEST; neighbor1 += 2)
    {
        neighbor2 = neighbor1 + 1;
        x_neighbor = get_x_neighbor(neighbor1, x, m);
        y_neighbor = get_y_neighbor(neighbor1, y, n);
        price_neighbor = price_min + *(strategies + x_neighbor * n + y_neighbor) * price_step_size;
        *(indifference_distances + neighbor1) = (price_neighbor - price + 1) / 2;
        x_neighbor = get_x_neighbor(neighbor2, x, m);
        y_neighbor = get_y_neighbor(neighbor2, y, n);
        price_neighbor = price_min + *(strategies + x_neighbor * n + y_neighbor) * price_step_size;
        *(indifference_distances + neighbor2) = (price_neighbor - price + 2) / 2;
    }
    market_share = 0;
    for (neighbor1 = NEIGHBOR_NORTH; neighbor1 <= NEIGHBOR_WEST; neighbor1 += 2)
    {
        neighbor2 = neighbor1 + 1;
        neighbor3 = (neighbor1 == NEIGHBOR_WEST) ? NEIGHBOR_NORTH : (neighbor1 + 2);
        market_share += *(indifference_distances + neighbor1) * *(indifference_distances + neighbor3);
        a = *(indifference_distances + neighbor1) - *(indifference_distances + neighbor2) + *(indifference_distances + neighbor3);
        if (a > 0)
            market_share -= a * a / 2;
    }
    return market_share;
}

double calc_market_share(int *strategies, int x, int y, int m, int n, double price_min, double price_step_size, int model_variant)
{
    return (model_variant == MODEL_VARIANT_A) ? calc_market_share_variant_A(strategies, x, y, m, n, price_min, price_step_size) : calc_market_share_variant_B(strategies, x, y, m, n, price_min, price_step_size);
}

void init_market_shares(double *market_shares, int *strategies, int m, int n, double price_min, double price_step_size, int model_variant)
{
    int x, y;

    for (x = 0; x < m; x++)
        for (y = 0; y < n; y++)
            *(market_shares + x * n + y) = calc_market_share(strategies, x, y, m, n, price_min, price_step_size, model_variant);
}

void calc_profits(double *profits, int *strategies, double *market_shares, int m, int n, double price_min, double price_step_size)
{
    double price;
    int x, y;

    for (x = 0; x < m; x++)
        for (y = 0; y < n; y++)
        {
            price = price_min + *(strategies + x * n + y) * price_step_size;
            *(profits + x * n + y) = price * *(market_shares + x * n + y);
        }
}

void update_market_shares(double *market_shares, int *strategies, int x, int y, int m, int n, double price_min, double price_step_size, int model_variant)
{
    int neighbor, x_neighbor, y_neighbor;

    *(market_shares + x * n + y) = calc_market_share(strategies, x, y, m, n, price_min, price_step_size, model_variant);
    for (neighbor = NEIGHBOR_NORTH; neighbor <= NEIGHBOR_NORTH_WEST; neighbor += (model_variant == MODEL_VARIANT_A) ? 2 : 1)
    {
        x_neighbor = get_x_neighbor(neighbor, x, m);
        y_neighbor = get_y_neighbor(neighbor, y, n);
        *(market_shares + x_neighbor * n + y_neighbor) = calc_market_share(strategies, x_neighbor, y_neighbor, m, n, price_min, price_step_size, model_variant);
    }
}

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    double *initial_strategies, *market_shares, *mean_strategy, *profits, *profits_per_strategy;
    double experimentation_prob, gset, mean_profit_new_strategy, mean_profit_strategy, noise_level, price_min, price_step_size;
    int *n_firms_per_strategy, *strategies;
    int i, iset, j, learning_neighborhood, m, model_variant, n, neighbor, new_strategy, n_rounds, n_strategies, strategy, strategy_neighbor, sum_strategies, x, x_neighbor, y, y_neighbor;
    long *iv;
    long idum, iy, random_seed;
    mwSize dims[2];

    initial_strategies = mxGetPr(prhs[0]);
    m = (int)mxGetN(prhs[0]);
    n = (int)mxGetM(prhs[0]);
    price_min = mxGetScalar(prhs[1]);
    price_step_size = mxGetScalar(prhs[2]);
    n_strategies = (int)mxGetScalar(prhs[3]);
    model_variant = (int)mxGetScalar(prhs[4]);
    learning_neighborhood = (int)mxGetScalar(prhs[5]);
    noise_level = mxGetScalar(prhs[6]);
    experimentation_prob = mxGetScalar(prhs[7]);
    n_rounds = (int)mxGetScalar(prhs[8]);
    random_seed = (long)mxGetScalar(prhs[9]);

    dims[0] = n;
    dims[1] = m;
    plhs[0] = mxCreateNumericArray(2, dims, mxINT32_CLASS, mxREAL);
    strategies = (int*)mxGetPr(plhs[0]);
    plhs[1] = mxCreateDoubleMatrix(n_rounds, 1, mxREAL);
    mean_strategy = mxGetPr(plhs[1]);

    idum = -labs(random_seed);
    iy = 0;
    iv = mxMalloc(NTAB * sizeof(*iv));
    iset = 0;

    init_strategies(strategies, &sum_strategies, initial_strategies, m, n);
    market_shares = mxMalloc(m * n * sizeof(*market_shares));
    init_market_shares(market_shares, strategies, m, n, price_min, price_step_size, model_variant);
    profits = mxMalloc(m * n * sizeof(*profits));
    profits_per_strategy = mxMalloc(n_strategies * sizeof(*profits_per_strategy));
    n_firms_per_strategy = mxMalloc(n_strategies * sizeof(*n_firms_per_strategy));

    for (i = 0; i < n_rounds; i++)
    {
        calc_profits(profits, strategies, market_shares, m, n, price_min, price_step_size);

        init_profits_per_strategy(profits_per_strategy, n_firms_per_strategy, n_strategies);
        x = (int)(m * ran1(&idum, &iy, iv)) % m;
        y = (int)(n * ran1(&idum, &iy, iv)) % n;
        strategy = *(strategies + x * n + y);
        *(profits_per_strategy + strategy) += *(profits + x * n + y);
        (*(n_firms_per_strategy + strategy))++;
        for (neighbor = NEIGHBOR_NORTH; neighbor <= NEIGHBOR_NORTH_WEST; neighbor += (learning_neighborhood == SMALL_LEARNING_NEIGHBORHOOD) ? 2 : 1)
        {
            x_neighbor = get_x_neighbor(neighbor, x, m);
            y_neighbor = get_y_neighbor(neighbor, y, n);
            strategy_neighbor = *(strategies + x_neighbor * n + y_neighbor);
            *(profits_per_strategy + strategy_neighbor) += *(profits + x_neighbor * n + y_neighbor) + noise_level * gasdev(&idum, &iy, iv, &iset, &gset);
            (*(n_firms_per_strategy + strategy_neighbor))++;
        }
        mean_profit_new_strategy = -DBL_MAX;
        for (j = 0; j < n_strategies; j++)
        {
            mean_profit_strategy = (*(n_firms_per_strategy + j) > 0) ? (*(profits_per_strategy + j) / *(n_firms_per_strategy + j) + 1e-10 * ran1(&idum, &iy, iv)) : -DBL_MAX;
            if (mean_profit_strategy > mean_profit_new_strategy)
            {
                new_strategy = j;
                mean_profit_new_strategy = mean_profit_strategy;
            }
        }
        if (new_strategy != strategy)
        {
            *(strategies + x * n + y) = new_strategy;
            sum_strategies += new_strategy - strategy;
            update_market_shares(market_shares, strategies, x, y, m, n, price_min, price_step_size, model_variant);
        }

        for (x = 0; x < m; x++)
            for (y = 0; y < n; y++)
                if (ran1(&idum, &iy, iv) < experimentation_prob)
                    if (ran1(&idum, &iy, iv) < 0.5)
                    {
                        if (*(strategies + x * n + y) > 0)
                        {
                            (*(strategies + x * n + y))--;
                            sum_strategies--;
                            update_market_shares(market_shares, strategies, x, y, m, n, price_min, price_step_size, model_variant);
                        }
                    }
                    else if (*(strategies + x * n + y) < n_strategies - 1)
                    {
                        (*(strategies + x * n + y))++;
                        sum_strategies++;
                        update_market_shares(market_shares, strategies, x, y, m, n, price_min, price_step_size, model_variant);
                    }

        *(mean_strategy + i) = (double)sum_strategies / (m * n);
    }

    mxFree(iv);
    mxFree(market_shares);
    mxFree(profits);
    mxFree(profits_per_strategy);
    mxFree(n_firms_per_strategy);
}
